@@ -30,7 +30,7 @@
 
 ;;; version.
 
-(defconst iiimcf-server-control-version "0.3 (Yodo)")
+(defconst iiimcf-server-control-version "0.4 (Inugami)")
 
 ;;; Customizable options.
 
@@ -152,22 +152,131 @@ If nil, don't apply input method preference to a server."
 (make-variable-buffer-local 'iiimcf-server-control-preedit-mode)
 (put 'iiimcf-server-control-preedit-mode 'permanent-local t)
 
+;; COM-ID base database entry.
+;;    (COM-ID . (entry-alist))
+(defvar iiimcf-server-control-private-com-id-database-alist nil)
+
+(defun iiimcf-server-control-register-com-id (com-id)
+  (if (null (assq com-id iiimcf-server-control-private-com-id-database-alist))
+      (setq iiimcf-server-control-private-com-id-database-alist
+	    (cons
+	     (list com-id)
+	     iiimcf-server-control-private-com-id-database-alist))))
+
+(defun iiimcf-server-control-unregister-com-id (com-id)
+  (let ((slot
+	 (assq com-id
+	       iiimcf-server-control-private-com-id-database-alist)))
+    (if slot
+	(setq iiimcf-server-control-private-com-id-database-alist
+	      (delq slot iiimcf-server-control-private-com-id-database-alist)))))
+
+(defun iiimcf-server-control-com-id-get (com-id key)
+  (cdr (assq key
+	     (cdr
+	      (assq com-id
+		    iiimcf-server-control-private-com-id-database-alist)))))
+
+(defun iiimcf-server-control-com-id-put (com-id key val)
+  (let ((slot
+	 (assq com-id
+	       iiimcf-server-control-private-com-id-database-alist))
+	sslot)
+    (if slot
+	(if (setq sslot (assq key (cdr slot)))
+	    (setcdr sslot val)
+	  (setcdr slot (cons (cons key val) (cdr slot))))
+      (error "COM-ID:%S is not registerd." com-id))))
+
+;; IM-ID base database entry.
+;;    (IM-ID . (entry-alist))
+(defvar iiimcf-server-control-private-im-id-database-alist nil)
+
+(defun iiimcf-server-control-register-im-id (im-id)
+  (if (null (assoc im-id iiimcf-server-control-private-im-id-database-alist))
+      (setq iiimcf-server-control-private-im-id-database-alist
+	    (cons
+	     (list im-id)
+	     iiimcf-server-control-private-im-id-database-alist))))
+
+(defun iiimcf-server-control-unregister-im-id (im-id)
+  (let ((slot
+	 (assoc im-id
+		iiimcf-server-control-private-im-id-database-alist)))
+    (if slot
+	(setq iiimcf-server-control-private-im-id-database-alist
+	      (delq slot iiimcf-server-control-private-im-id-database-alist)))))
+
+(defun iiimcf-server-control-im-id-get (im-id key)
+  (cdr (assq key
+	     (cdr
+	      (assoc im-id
+		     iiimcf-server-control-private-im-id-database-alist)))))
+
+(defun iiimcf-server-control-im-id-put (im-id key val)
+  (let ((slot
+	 (assoc im-id
+		iiimcf-server-control-private-im-id-database-alist))
+	sslot)
+    (if slot
+	(if (setq sslot (assq key (cdr slot)))
+	    (setcdr sslot val)
+	  (setcdr slot (cons (cons key val) (cdr slot))))
+      (error "IM-ID:%S is not registerd." im-id))))
+
+;; IC-ID base database entry.
+;;    (IC-ID . (entry-alist))
+(defvar iiimcf-server-control-private-ic-id-database-alist nil)
+
+(defun iiimcf-server-control-register-ic-id (ic-id)
+  (if (null (assoc ic-id iiimcf-server-control-private-ic-id-database-alist))
+      (setq iiimcf-server-control-private-ic-id-database-alist
+	    (cons
+	     (list ic-id)
+	     iiimcf-server-control-private-ic-id-database-alist))))
+
+(defun iiimcf-server-control-unregister-ic-id (ic-id)
+  (let ((slot
+	 (assoc ic-id
+		iiimcf-server-control-private-ic-id-database-alist)))
+    (if slot
+	(setq iiimcf-server-control-private-ic-id-database-alist
+	      (delq slot iiimcf-server-control-private-ic-id-database-alist)))))
+
+(defun iiimcf-server-control-ic-id-get (ic-id key)
+  (cdr (assq key
+	     (cdr
+	      (assoc ic-id
+		     iiimcf-server-control-private-ic-id-database-alist)))))
+
+(defun iiimcf-server-control-ic-id-put (ic-id key val)
+  (let ((slot
+	 (assoc ic-id
+		iiimcf-server-control-private-ic-id-database-alist))
+	sslot)
+    (if slot
+	(if (setq sslot (assq key (cdr slot)))
+	    (setcdr sslot val)
+	  (setcdr slot (cons (cons key val) (cdr slot))))
+      (error "IC-ID:%S is not registerd." ic-id))))
+
 (defvar iiimcf-server-control-com-id nil)
 (defvar iiimcf-server-control-im-id nil)
 (defvar iiimcf-server-control-ic-id nil)
+(make-variable-buffer-local 'iiimcf-server-control-ic-id)
+(put 'iiimcf-server-control-ic-id 'permanent-local t)
 (defvar iiimcf-server-control-dynamic-event-flow-p nil)
 (defvar iiimcf-server-control-editing-buffer nil
   "The buffer where a user is inputting currently.")
-(defvar iiimcf-server-control-preedit "")
-(defvar iiimcf-server-control-preedit-marker (make-marker))
 ;; (defvar iiimcf-server-control-caret-overlay
 ;;   (make-overlay 0 0))
 ;; (overlay-put iiimcf-server-control-caret-overlay
 ;;	     'face 'modeline)
-(defvar iiimcf-server-control-maintaining-preedit-text nil)
 
 (defvar iiimcf-server-control-mode-line "")
 (make-variable-buffer-local 'iiimcf-server-control-mode-line)
+(defvar iiimcf-server-control-status "")
+(make-variable-buffer-local 'iiimcf-server-control-status)
 (if (null (memq 'iiimcf-server-control-mode-line mode-line-format))
     (let ((ll mode-line-format)
 	  gen elem)
@@ -188,32 +297,6 @@ If nil, don't apply input method preference to a server."
 (defvar iiimcf-server-control-current-lookup-choice nil)
 (defvar iiimcf-server-control-lookup-choice-configuration nil)
 
-;;; private database
-
-(defvar iiimcf-server-control-com-id-plists nil)
-
-(defun iiimcf-server-control-register-com-id (com-id)
-  (if (null (assq com-id iiimcf-server-control-com-id-plists))
-      (setq iiimcf-server-control-com-id-plists
-	    (cons (list com-id)
-		  iiimcf-server-control-com-id-plists))))
-
-(defun iiimcf-server-control-unregister-com-id (com-id)
-  (let ((slot (assq com-id iiimcf-server-control-com-id-plists)))
-    (if slot
-	(setq iiimcf-server-control-com-id-plists
-	      (delq slot
-		    iiimcf-server-control-com-id-plists)))))
-
-(defsubst iiimcf-server-control-com-id-get (com-id prop)
-  (plist-get (cdr (assq com-id iiimcf-server-control-com-id-plists))
-	     prop))
-
-(defsubst iiimcf-server-control-com-id-put (com-id prop val)
-  (let ((slot (assq com-id iiimcf-server-control-com-id-plists)))
-    (if slot
-	(setcdr slot (plist-put (cdr slot) prop val)))))
-
 ;;; keymap
 
 ;; keymap for initial state.
@@ -350,6 +433,25 @@ If nil, don't apply input method preference to a server."
      ;; (if im (list (cons 'input-method im)) nil)
      )))
 
+(defun iiimcf-server-control-im-setup (com-id)
+  (let (mes im-id)
+    (iiimp-send-message
+     'iiimp-im-connect
+     com-id
+     iiimcf-server-control-username)
+    (setq mes (iiimcf-message-manager
+	       com-id
+	       (list
+		(list 'iiimp-im-connect-reply com-id))))
+    (setq im-id (nth 2 mes))
+    (iiimcf-server-control-register-im-id
+     im-id)
+    (iiimcf-send-client-descriptor
+     com-id im-id
+     (format "IIIMCF-SC/%s"
+	     iiimcf-server-control-version))
+    im-id))
+
 (defun iiimcf-server-control-ic-setup (com-id im-id)
   (let (ic-id mes)
     (iiimp-send-message
@@ -363,7 +465,9 @@ If nil, don't apply input method preference to a server."
 	   (list
 	    (list 'iiimp-im-createic-reply
 		  com-id im-id))))
-    (setq ic-id (nth 3 mes))
+    (setq ic-id (iiimp-message-ic-id mes))
+    (iiimcf-server-control-register-ic-id
+     ic-id)
     (iiimp-send-message 'iiimp-im-resetic com-id im-id ic-id)
     (iiimcf-message-manager
      com-id
@@ -372,6 +476,15 @@ If nil, don't apply input method preference to a server."
 	    com-id im-id ic-id)))
     ic-id))
 
+(defun iiimcf-server-control-im-destruct (com-id im-id)
+  (iiimp-send-message
+   'iiimp-im-disconnect
+   com-id im-id)
+  (iiimcf-message-manager
+   com-id
+   (list (list 'iiimp-im-disconnect-reply com-id im-id)))
+  (iiimcf-server-control-unregister-im-id im-id))
+
 (defun iiimcf-server-control-ic-destruct (com-id im-id ic-id)
   (iiimp-send-message 'iiimp-im-resetic com-id im-id ic-id)
   (iiimcf-message-manager
@@ -384,7 +497,8 @@ If nil, don't apply input method preference to a server."
    com-id
    (list
     (list 'iiimp-im-destroyic-reply
-	  com-id im-id ic-id))))
+	  com-id im-id ic-id)))
+  (iiimcf-server-control-unregister-ic-id ic-id))
 
 (defmacro iiimcf-server-control-ic-operation
   (x-com-id x-im-id x-ic-id &rest form)
@@ -427,37 +541,18 @@ If nil, don't apply input method preference to a server."
 	      (car iiimcf-server-control-hostlist)))))
   (iiimcf-server-control-register-com-id
    iiimcf-server-control-com-id)
-  (let ((com-id iiimcf-server-control-com-id)
-	mes)
+  (let ((com-id iiimcf-server-control-com-id))
     (iiimcf-server-control-register-handlers
      iiimcf-server-control-com-level-handler-alist
      com-id)
     (if (null iiimcf-server-control-im-id)
-	(progn
-	  (iiimp-send-message
-	   'iiimp-im-connect
-	   com-id
-	   iiimcf-server-control-username)
-	  (setq mes (iiimcf-message-manager
-		     com-id
-		     (list
-		      (list 'iiimp-im-connect-reply com-id))))
-	  (setq iiimcf-server-control-im-id
-		(nth 2 mes))
-	  (iiimcf-send-client-descriptor
-	   com-id iiimcf-server-control-im-id
-	   (format "IIIMCF-SC/%s"
-		   iiimcf-server-control-version))))
-    (if (null iiimcf-server-control-ic-id)
-	(setq iiimcf-server-control-ic-id
-	      (iiimcf-server-control-ic-setup
-	       com-id iiimcf-server-control-im-id)))
+	(setq iiimcf-server-control-im-id
+	      (iiimcf-server-control-im-setup com-id)))
 
     (iiimcf-server-control-register-handlers
      iiimcf-server-control-ic-level-handler-alist
      com-id
-     iiimcf-server-control-im-id
-     iiimcf-server-control-ic-id)
+     iiimcf-server-control-im-id)
 
     (if iiimcf-server-control-async-invocation-p
 	(iiimp-enable-async-invocation
@@ -480,17 +575,9 @@ If nil, don't apply input method preference to a server."
 	(setq iiimcf-server-control-ic-id nil)))
   (if iiimcf-server-control-im-id
       (progn
-	(iiimp-send-message
-	 'iiimp-im-disconnect
+	(iiimcf-server-control-im-destruct
 	 iiimcf-server-control-com-id
 	 iiimcf-server-control-im-id)
-	(iiimcf-message-manager
-	 iiimcf-server-control-com-id
-	 (list
-	  (list
-	   'iiimp-im-disconnect-reply
-	   iiimcf-server-control-com-id
-	   iiimcf-server-control-im-id)))
 	(setq iiimcf-server-control-im-id nil)))
   (if iiimcf-server-control-com-id
       (progn
@@ -512,20 +599,16 @@ If nil, don't apply input method preference to a server."
   (interactive)
   (iiimcf-server-control-activate -1))
 
-(defun iiimcf-server-control-notify-trigger (flag)
+(defun iiimcf-server-control-notify-trigger (com-id im-id ic-id flag)
   (iiimp-send-message
    'iiimp-im-trigger-notify
-   iiimcf-server-control-com-id
-   iiimcf-server-control-im-id
-   iiimcf-server-control-ic-id
+   com-id im-id ic-id
    flag)
   (iiimcf-message-manager
-   iiimcf-server-control-com-id
+   com-id
    (list
     (list 'iiimp-im-trigger-notify-reply
-	  iiimcf-server-control-com-id
-	  iiimcf-server-control-im-id
-	  iiimcf-server-control-ic-id))))
+	  com-id im-id ic-id))))
 
 (defun iiimcf-server-control-remove-minor-mode-map ()
   (let ((islot (assq 'iiimcf-server-control-enable
@@ -556,23 +639,34 @@ If nil, don't apply input method preference to a server."
       (unwind-protect
 	  (progn
 	    (setq describe-current-input-method-function nil)
-	    (if (iiimp-check-channel-connection
-		 iiimcf-server-control-com-id)
-		(iiimcf-server-control-notify-trigger nil))
-	    (iiimcf-server-control-clear-preedit-text)
+	    (if (and (iiimp-check-channel-connection
+		      iiimcf-server-control-com-id)
+		     iiimcf-server-control-ic-id)
+		(iiimcf-server-control-notify-trigger
+		 iiimcf-server-control-com-id
+		 iiimcf-server-control-im-id
+		 iiimcf-server-control-ic-id
+		 nil))
 	    ;;(run-hooks 'iiimcf-server-control-inactivate-hook)
 	    )
 	(setq iiimcf-server-control-enable nil)
+	(setq iiimcf-server-control-mode-line "")
 	(iiimcf-server-control-remove-minor-mode-map))
 
     ;; activate
     (setq iiimcf-server-control-enable t)
     (iiimcf-server-control-setup)
-    (iiimcf-server-control-notify-trigger t)
     (setq inactivate-current-input-method-function
 	  'iiimcf-server-control-inactivate)
     (setq describe-current-input-method-function
 	  'iiimcf-server-control-help)
+    ;; setup IC, mainly for status information.
+    (iiimcf-server-control-prepare-current-ic-id
+     iiimcf-server-control-com-id
+     iiimcf-server-control-im-id)
+    (setq iiimcf-server-control-mode-line
+	  (format "[%s]" iiimcf-server-control-status))
+    (force-mode-line-update)
 
     ;; inactivate the current input method also in minibuffers
     ;; before exiting.
@@ -590,53 +684,8 @@ If nil, don't apply input method preference to a server."
 		   (function
 		    iiimcf-server-control-exit-from-minibuffer))))
 
-(defun iiimcf-server-control-keyforward (&optional ev)
-  (interactive)
-  (setq iiimcf-server-control-editing-buffer
-	(current-buffer))
-  (let* ((event (or ev
-		    last-command-event))
-	 (keyevent (iiimcf-translate-emacs-event event)))
-    ;;(if (eq event ? ) (setq event 'convert))
-    (if keyevent
-	(condition-case err
-	    (progn
-	      (iiimp-send-message
-	       'iiimp-im-forward-event
-	       iiimcf-server-control-com-id
-	       iiimcf-server-control-im-id
-	       iiimcf-server-control-ic-id
-	       (list 'keyevent
-		     (list keyevent)))
-	      ;; I have no idea why Solaris 8 IIIM server
-	      ;; does not respond to IM_FORWARD_EVENT sometimes.
-	      (iiimcf-message-manager
-	       iiimcf-server-control-com-id
-	       (list
-		(list 'iiimp-im-forward-event-reply
-		      iiimcf-server-control-com-id
-		      iiimcf-server-control-im-id
-		      iiimcf-server-control-ic-id)
-		(list 'iiimp-im-forward-event
-		      iiimcf-server-control-com-id
-		      iiimcf-server-control-im-id
-		      iiimcf-server-control-ic-id))))
-	  (iiimp-fatal
-	   (iiimcf-server-control-shutdown)
-	   (signal (car err) (cdr err))))
-      ;; Don't send an unattended event to server,
-      ;; Deal with it by myself.
-      (iiimcf-server-control-dispatch-emacs-event (list event)))
-    nil))
-
-(defun iiimcf-server-control-trigger-notify (mes)
-  (let ((flag (nth 4 mes)))
-    (if flag
-	nil
-      (inactivate-input-method))))
-
 (defun iiimcf-server-control-forward-event (mes)
-  (let ((con (nth 4 mes))
+  (let ((con (iiimp-message-forward-event-contents mes))
 	kevslot1 kevslot2 kev)
     (iiimp-add-debug-log
      (format "FWD:%S\n" con))
@@ -663,7 +712,7 @@ If nil, don't apply input method preference to a server."
 
 (defun iiimcf-server-control-completing-read
   (prompt annotated-collection)
-  (let ((completion-setup-hook))
+  (let ((completion-setup-hook completion-setup-hook))
     (add-hook 'completion-setup-hook
 	      (lambda ()
 		(let ((inhibit-read-only t)
@@ -687,8 +736,11 @@ If nil, don't apply input method preference to a server."
 
 (defun iiimcf-server-control-maintain-language-list (mes)
   (iiimcf-server-control-com-id-put
-   (nth 1 mes) 'language-list (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   'language-list
+   (iiimp-message-im-language-list mes)))
 
+;;; FIXME
 (defun iiimcf-server-control-change-language ()
   (interactive)
   (inactivate-input-method)
@@ -701,22 +753,16 @@ If nil, don't apply input method preference to a server."
 	  (iiimcf-server-control-com-id-get
 	   iiimcf-server-control-com-id
 	   'language-list))))
-  (iiimcf-server-control-ic-destruct
-   iiimcf-server-control-com-id
-   iiimcf-server-control-im-id
-   iiimcf-server-control-ic-id)
-  (setq iiimcf-server-control-ic-id
-	(iiimcf-server-control-ic-setup
-	 iiimcf-server-control-com-id
-	 iiimcf-server-control-im-id))
   (activate-input-method 'iiim-server-control))
 
 (defun iiimcf-server-control-maintain-input-method-list (mes)
   (iiimcf-server-control-com-id-put
-   (nth 1 mes)
+   (iiimp-message-com-id mes)
    'input-method-list
-   (nth 1 (assq 'input-method-list (nth 3 mes)))))
+   (nth 1 (assq 'input-method-list
+		(iiimp-message-im-attribute-list mes)))))
 
+;;; FIXME
 (defun iiimcf-server-control-change-input-method ()
   (interactive)
   (inactivate-input-method)
@@ -730,22 +776,83 @@ If nil, don't apply input method preference to a server."
 	   iiimcf-server-control-com-id
 	   'input-method-list))
 	  nil t))
-  (iiimcf-server-control-ic-destruct
-   iiimcf-server-control-com-id
-   iiimcf-server-control-im-id
-   iiimcf-server-control-ic-id)
-  (setq iiimcf-server-control-ic-id
-	(iiimcf-server-control-ic-setup
-	 iiimcf-server-control-com-id
-	 iiimcf-server-control-im-id))
   (activate-input-method 'iiim-server-control))
 
+;;;
+;;; trigger-notify operations
+;;;
+
+(defun iiimcf-server-control-trigger-notify (mes)
+  (let ((flag (iiimp-message-trigger-notify-flag mes)))
+    (if flag
+	nil
+      (inactivate-input-method))))
+
 ;;;
 ;;; Preedit operations
 ;;;
 
+;;
+;; XX ... preedit-open-string
+;; YY ... preedit-close-string
+;;
+;;    +---- iiimcf-server-control-preedit
+;;  +-+------+
+;;  v        v
+;;  XX abcd YY
+;;          ^^-- iiimcf-server-control-preedit-end, rear-nonsticky
+;;
+
 (defsubst iiimcf-server-control-preedit-enable-p ()
-  (> (length iiimcf-server-control-preedit) 0))
+  (get-text-property (point) 'iiimcf-server-control-preedit))
+
+(defsubst iiimcf-server-control-current-context ()
+  (get-text-property (point) 'iiimcf-server-control-ids))
+
+(defsubst iiimcf-server-control-prepare-current-ic-id (com-id im-id)
+  (let ((ic-id
+	 (or (nth 2 (iiimcf-server-control-current-context))
+	     iiimcf-server-control-ic-id
+	     (setq iiimcf-server-control-ic-id
+		   (iiimcf-server-control-ic-setup
+		    com-id im-id)))))
+    (iiimcf-server-control-notify-trigger
+     com-id im-id ic-id t)
+    ic-id))
+
+(defun iiimcf-server-control-enter-or-leave-preedit
+  (oldpos newpos)
+  (let ((oldids (get-text-property oldpos 'iiimcf-server-control-ids))
+	(newids (get-text-property newpos 'iiimcf-server-control-ids)))
+    (if oldids
+	;; inactivate ic and keymap.
+	(progn
+	  ;;(iiimcf-server-control-notify-trigger
+	  ;;(car oldids)
+	  ;;(nth 1 oldids)
+	  ;;(nth 2 oldids)
+	  ;; nil)
+	  (setq iiimcf-server-control-preedit-mode nil
+		iiimcf-server-control-ic-id nil)))
+    (if newids
+	;; activate ic and keymap.
+	(progn
+	  (iiimcf-server-control-notify-trigger
+	   (car newids)
+	   (nth 1 newids)
+	   (nth 2 newids)
+	   t)
+	  ;; destruct old ic if exist.
+	  (if (and iiimcf-server-control-ic-id
+		   (not (eq (nth 2 newids)
+			    iiimcf-server-control-ic-id)))
+	      (iiimcf-server-control-ic-destruct
+	       iiimcf-server-control-com-id
+	       iiimcf-server-control-im-id
+	       iiimcf-server-control-ic-id))
+	  (setq iiimcf-server-control-enable t
+		iiimcf-server-control-preedit-mode t
+		iiimcf-server-control-ic-id (nth 2 newids))))))
 
 (defun iiimcf-server-control-put-face-to-preedit (str)
   (let ((len (length str))
@@ -810,43 +917,43 @@ If nil, don't apply input method preference to a server."
     (iiimcf-server-control-enclose-preedit-with-string
      str caret)))
 
-(defun iiimcf-server-control-setup-preedit-text ()
-  (let ((prop-begin
-	 '(iiimcf-server-control-preedit-start t intangible t))
+(defun iiimcf-server-control-setup-preedit-text (com-id im-id ic-id marker)
+  (let ((props
+	 (list 'iiimcf-server-control-preedit t
+	       'intangible t
+	       'read-only t
+	       'point-entered 'iiimcf-server-control-enter-or-leave-preedit
+	       'point-left 'iiimcf-server-control-enter-or-leave-preedit
+	       'iiimcf-server-control-ids
+	       (list com-id im-id ic-id)))
 	(prop-end
-	 '(iiimcf-server-control-preedit-end t intangible t
-	   rear-nonsticky t start-open t))
+	 '(iiimcf-server-control-preedit-end t
+           intangible t rear-nonsticky t start-open t))
 	(oldbuf (current-buffer))
 	pss pse)
-    ;; (if (string= " *IIIMP" (buffer-name oldbuf))
-    ;;    (debug))
+    (if (null marker) (setq marker (make-marker)))
     (set-buffer iiimcf-server-control-editing-buffer)
     (setq pss (point))
     (let ((buffer-undo-list t))
       (insert iiimcf-server-control-preedit-open-string)
       (setq pse (point))
-      (add-text-properties pss pse prop-begin) 
       (insert iiimcf-server-control-preedit-close-string)
       (add-text-properties pse (point) prop-end)
-      (put-text-property pss (point) 'read-only t)
-      (set-marker iiimcf-server-control-preedit-marker
-		  pss)
-      (set-marker-insertion-type
-       iiimcf-server-control-preedit-marker t))
-    (set-buffer oldbuf)))
+      (add-text-properties pss (point) props) 
+      (set-marker marker pss)
+      (set-marker-insertion-type marker t))
+    (set-buffer oldbuf)
+    (iiimcf-server-control-ic-id-put
+     ic-id 'preedit-marker marker)
+    marker))
 
-(defun iiimcf-server-control-insert-preedit (text caret)
-  (if (null (marker-position
-	     iiimcf-server-control-preedit-marker))
-      (iiimcf-server-control-setup-preedit-text))
+(defun iiimcf-server-control-insert-preedit (text caret marker)
   (let ((inhibit-read-only t)
 	(inhibit-point-motion-hooks t)
 	(oldbuf (current-buffer))
-	(pts (marker-position
-	      iiimcf-server-control-preedit-marker))
+	(pts (marker-position marker))
 	pte)
-    (set-buffer (marker-buffer
-		 iiimcf-server-control-preedit-marker))
+    (set-buffer (marker-buffer marker))
     (setq pte
 	  (next-single-property-change
 	   pts 'iiimcf-server-control-preedit-end))
@@ -863,9 +970,7 @@ If nil, don't apply input method preference to a server."
 
 (defun iiimcf-server-control-show-preedit-minibuffer-p ()
   (and input-method-use-echo-area
-       (null
-	(marker-position
-	 iiimcf-server-control-preedit-marker))))
+       (iiimcf-server-control-preedit-enable-p)))
 
 (defun iiimcf-server-control-show-preedit-in-minibuffer (disptext caret)
   (let (pt mbuf)
@@ -883,59 +988,76 @@ If nil, don't apply input method preference to a server."
       ;  (+ pt caret) (+ 1 pt caret) mbuf)
       )))
 
-(defun iiimcf-server-control-maintain-preedit-text-internal
-  (mes &optional init)
-  (let* ((caret (nth 4 mes))
-	 (ch-first (nth 5 mes))
-	 (ch-len (nth 6 mes))
-	 (ch-second (+ ch-first ch-len))
-	 (contents (nth 7 mes))
+(defun iiimcf-server-control-update-preedit-text (mes preedit-text)
+  (let* ((data (iiimp-message-preedit-draw-data mes))
+	 (caret (aref data 0))
+	 (ch-first (aref data 1))
+	 (ch-len (aref data 2))
+	 (contents (aref data 3))
 	 (str (nth 1 contents))
-	 (len (length iiimcf-server-control-preedit))
-	 disptext head tail)
+	 (ch-second (+ ch-first ch-len))
+	 (len (length preedit-text))
+	 head tail)
     (if (and (>= len ch-first)
 	     (> ch-first 0))
 	(setq head (substring
-		    iiimcf-server-control-preedit
+		    preedit-text
 		    0 ch-first)))
     (if (> len ch-second)
 	(setq tail (substring
-		    iiimcf-server-control-preedit
+		    preedit-text
 		    ch-second)))
-    (setq iiimcf-server-control-preedit (concat head str tail)
-	  disptext (iiimcf-server-control-format-preedit
-		    iiimcf-server-control-preedit caret)
-	  caret (cdr disptext)
-	  disptext (car disptext))
+    (setq preedit-text (concat head str tail))
+    (iiimcf-server-control-format-preedit
+     preedit-text caret)))
 
-    (if (iiimcf-server-control-preedit-enable-p)
+(defun iiimcf-server-control-maintain-preedit-text-internal (mes)
+  (let* ((com-id (iiimp-message-com-id mes))
+	 (im-id (iiimp-message-im-id mes))
+	 (ic-id (iiimp-message-ic-id mes))
+	 (preedit-text (or (iiimcf-server-control-ic-id-get
+			    ic-id 'preedit-text)
+			   ""))
+	 (caret (iiimcf-server-control-update-preedit-text
+		 mes preedit-text))
+	 marker)
+    (setq preedit-text (car caret)
+	  caret (cdr caret))
+    (iiimcf-server-control-ic-id-put
+     ic-id 'preedit-text preedit-text)
+    (if (> (length preedit-text) 0)
 	;; Show preedit text.
 	(if (iiimcf-server-control-show-preedit-minibuffer-p)
 	    (iiimcf-server-control-show-preedit-in-minibuffer
-	     disptext caret)
-	  (if init
-	      (iiimcf-server-control-setup-preedit-text))
-	  (iiimcf-server-control-insert-preedit disptext caret))
-      (iiimcf-server-control-clear-preedit-text))))
+	     preedit-text caret)
+	  (setq marker
+		(iiimcf-server-control-ic-id-get
+		 ic-id 'preedit-marker))
+	  (if (not (and marker
+			(marker-position marker)))
+	      (setq marker
+		    (iiimcf-server-control-setup-preedit-text
+		     com-id im-id ic-id marker)))
+	  (iiimcf-server-control-insert-preedit
+	   preedit-text caret marker))
+      (iiimcf-server-control-clear-preedit-text
+       com-id im-id ic-id))))
 
 (defun iiimcf-server-control-maintain-preedit-text (mes)
   (let (;; disable the font-lock function invoked by
 	;; after-change-functions during preedit.
 	;; Even read-only text-property cannot protect inserted preedit
 	;; text from modifications of the face text-property by font-lock.
-	(font-lock-fontify-region-function (function ignore)))
-    (if (iiimcf-server-control-preedit-enable-p)
-	(iiimcf-server-control-maintain-preedit-text-internal mes)
-      ;; initialize preedit text.
-      (iiimcf-server-control-maintain-preedit-text-internal mes t)))
-  ;; set iiimcf-server-control-preedit-mode flag
-  (save-excursion
-    (set-buffer iiimcf-server-control-editing-buffer)
-    (setq iiimcf-server-control-preedit-mode
-	  (iiimcf-server-control-preedit-enable-p))))
+	(font-lock-fontify-region-function (function ignore))
+	(inhibit-point-motion-hooks t))
+    (iiimcf-server-control-maintain-preedit-text-internal mes)
+    ;; set iiimcf-server-control-preedit-mode flag
+    (save-excursion
+      (set-buffer iiimcf-server-control-editing-buffer)
+      (setq iiimcf-server-control-preedit-mode
+	    (iiimcf-server-control-preedit-enable-p)))))
 
-(defun iiimcf-server-control-clear-preedit-text ()
-  (setq iiimcf-server-control-preedit "")
+(defun iiimcf-server-control-clear-preedit-text (com-id im-id ic-id)
   (if (iiimcf-server-control-show-preedit-minibuffer-p)
       ;; erase minibuffer.
       (save-excursion
@@ -947,27 +1069,31 @@ If nil, don't apply input method preference to a server."
 	  ;; 0 0 mbuf)
 	  ))
     ;; Erase inserted preedit.
-      (let ((pts (marker-position
-		  iiimcf-server-control-preedit-marker))
-	    (inhibit-read-only t)
-	    pte)
-	(if pts
-	    (progn
-	      (save-excursion
-		(set-buffer (marker-buffer
-			     iiimcf-server-control-preedit-marker))
-		(setq pte
-		      (next-single-property-change
-		       pts 'iiimcf-server-control-preedit-end))
-		(if (null pte)
-		    (error "PREEDIT TEXT BROKEN!!"))
-		(let ((buffer-undo-list t))
-		  (delete-region
-		   pts (+ pte
-			  (length
-			   iiimcf-server-control-preedit-close-string)))))
-	      (goto-char pts)
-	      (set-marker iiimcf-server-control-preedit-marker nil))))))
+    (let* ((marker (iiimcf-server-control-ic-id-get
+		    ic-id 'preedit-marker))
+	   (pts (if marker (marker-position marker)))
+	   (inhibit-read-only t)
+	   pte)
+      (if pts
+	  (progn
+	    (save-excursion
+	      (set-buffer (marker-buffer marker))
+	      (setq pte
+		    (next-single-property-change
+		     pts 'iiimcf-server-control-preedit-end))
+	      (if (null pte)
+		  (error "PREEDIT TEXT IS BROKEN!!"))
+	      (let ((buffer-undo-list t))
+		(delete-region
+		 pts (+ pte
+			(length
+			 iiimcf-server-control-preedit-close-string)))))
+	    (goto-char pts)
+	    (set-marker marker nil))))))
+
+;;;
+;;; Lookup-choice operation.
+;;;
 
 (defun iiimcf-server-control-prepare-lookup-choice (mes)
   (setq iiimcf-server-control-lookup-choice-configuration mes))
@@ -978,18 +1104,21 @@ If nil, don't apply input method preference to a server."
   (with-output-to-temp-buffer "*Lookup choice*"
     (save-excursion
      (set-buffer standard-output)
-     (let ((i 0)
-	   (k 0)
-	   (num (nth 5 iiimcf-server-control-lookup-choice-configuration))
-	   (rows (nth 6 iiimcf-server-control-lookup-choice-configuration))
-	   (cols (nth 7 iiimcf-server-control-lookup-choice-configuration))
-	   (idxfirst (nth 4 mes))
-	   (idxlast (nth 5 mes))
-	   (idxcur (nth 6 mes))
-	   (candlist (nth 7 mes))
-	   (lblist (nth 8 mes))
-	   (title (nth 9 mes))
-	   curcand curlb)
+     (let* ((i 0)
+	    (k 0)
+	    (lsdata (iiimp-message-lookup-choice-start-data
+		     iiimcf-server-control-lookup-choice-configuration))
+	    (lcdata (iiimp-message-lookup-choice-draw-data mes))
+	    (num (aref lsdata 1))
+	    (rows (aref lsdata 2))
+	    (cols (aref lsdata 3))
+	    (idxfirst (aref lcdata 0))
+	    (idxlast (aref lcdata 1))
+	    (idxcur (aref lcdata 2))
+	    (candlist (aref lcdata 3))
+	    (lblist (aref lcdata 4))
+	    (title (aref lcdata 5))
+	    curcand curlb)
        (setq iiimcf-server-control-current-lookup-choice
 	     (list idxfirst candlist lblist title idxfirst idxlast))
        (insert (format "Title:%s\n" title))
@@ -1061,33 +1190,62 @@ If nil, don't apply input method preference to a server."
 	 (win (get-buffer-window buf)))
     (if win (delete-window win))))
 
+;;;
+;;; draw status operation.
+;;;
+
 (defun iiimcf-server-control-draw-status (mes)
-  (setq iiimcf-server-control-mode-line
-	(format "[%s]" (nth 1 (nth 4 mes))))
-  (force-mode-line-update))
+  (let ((ic-id (iiimp-message-ic-id mes))
+	marker buf)
+    (setq marker (iiimcf-server-control-ic-id-get
+		  ic-id 'preedit-marker))
+    (if marker
+	(setq buf (marker-buffer marker)))
+    (save-excursion
+      (if buf (set-buffer buf))
+      (setq iiimcf-server-control-status
+	    (nth 1 (iiimp-message-status-draw-contents mes)))
+      (setq iiimcf-server-control-mode-line
+	    (format "[%s]" iiimcf-server-control-status))
+      (force-mode-line-update))))
+
+;;;
+;;; draw AUX data operation.
+;;;
 
 (defun iiimcf-server-control-draw-aux-data (mes)
-  (let ((name (nth 5 mes))
-	(strs (nth 7 mes)))
+  (let ((name (iiimp-message-aux-im-name mes))
+	(strs (iiimp-message-aux-string-list mes)))
     (cond ((string-match "^jp\\.co\\.justsystem\\..*LookupAux$"
 			 name)
 	   (if strs
-	       (message "%s:  %s"
-			(car strs)
-			(mapconcat 
-			 (function identity)
-			 (cdr strs) " "))
-	     (message ""))))))
+	       (let ((i 1)
+		     (s (cdr strs))
+		     (r (format "%s:" (car strs))))
+		 (while s
+		   (setq r (concat r (format " %d:%s" i (car s)))
+			 i (1+ i)
+			 s (cdr s)))
+		 (princ r t))
+	     ;;(message "")
+	     )))))
   
+;;;
+;;; commit string operation.
+;;;
 
 (defun iiimcf-server-control-commit-string (mes)
-  (let ((str (copy-sequence (nth 1 (nth 4 mes))))
-	(oldbuf (current-buffer))
-	(buf (marker-buffer
-	      iiimcf-server-control-preedit-marker))
-	(pt (marker-position
-	     iiimcf-server-control-preedit-marker))
-	generated-events)
+  (let* ((ic-id (iiimp-message-ic-id mes))
+	 (marker (iiimcf-server-control-ic-id-get
+		  ic-id 'preedit-marker))
+	 (str (copy-sequence
+	       (nth 1 (iiimp-message-committed-string mes))))
+	 (oldbuf (current-buffer))
+	 (buf (marker-buffer marker))
+	 (pt (marker-position marker))
+	 (inhibit-point-motion-hooks t)
+	 generated-events)
+    ;; strip text properties from str.
     (set-text-properties 0 (length str) nil str)
     (if pt
 	(progn
@@ -1098,12 +1256,61 @@ If nil, don't apply input method preference to a server."
     ;; dispatch them immediately!
     (iiimcf-server-control-dispatch-emacs-event generated-events)
     ;; (insert str)
-    (if pt (set-marker
-	    iiimcf-server-control-preedit-marker
-	    (point)))
+    (if pt (set-marker marker (point)))
     (set-buffer oldbuf)
     (run-hooks 'input-method-after-insert-chunk-hook)))
 
+;;;
+;;; event handling routines (invoked with the keymaps).
+;;;
+
+;;  Forward keyevents to a IIIM server.
+(defun iiimcf-server-control-keyforward (&optional ev)
+  (interactive)
+  (setq iiimcf-server-control-editing-buffer
+	(current-buffer))
+  (let* ((event (or ev
+		    last-command-event))
+	 (keyevent (iiimcf-translate-emacs-event event))
+	 (ic-id (iiimcf-server-control-prepare-current-ic-id
+		 iiimcf-server-control-com-id
+		 iiimcf-server-control-im-id)))
+    ;;(if (eq event ? ) (setq event 'convert))
+    (if keyevent
+	(condition-case err
+	    (progn
+	      (iiimp-send-message
+	       'iiimp-im-forward-event
+	       iiimcf-server-control-com-id
+	       iiimcf-server-control-im-id
+	       ic-id
+	       (list 'keyevent
+		     (list keyevent)))
+	      ;; I have no idea why Solaris 8 IIIM server
+	      ;; does not respond to IM_FORWARD_EVENT sometimes.
+	      (iiimcf-message-manager
+	       iiimcf-server-control-com-id
+	       (list
+		(list 'iiimp-im-forward-event-reply
+		      iiimcf-server-control-com-id
+		      iiimcf-server-control-im-id
+		      ic-id)
+		(list 'iiimp-im-forward-event
+		      iiimcf-server-control-com-id
+		      iiimcf-server-control-im-id
+		      ic-id))))
+	  (iiimp-fatal
+	   (iiimcf-server-control-shutdown)
+	   (signal (car err) (cdr err))))
+      ;; Don't send an unattended event to server,
+      ;; Deal with it by myself.
+      (iiimcf-server-control-dispatch-emacs-event (list event)))
+    nil))
+
+;;;
+;;; Register input method.
+;;;
+
 (register-input-method
  "iiim-server-control" "Japanese"
  'iiimcf-server-control-activate
