@@ -36,7 +36,7 @@
 
 ;;; version.
 
-(defconst iiimcf-version "0.3 (Tenpozan)")
+(defconst iiimcf-version "0.4 (Oikedake)")
 
 ;;; error handling.
 
@@ -753,7 +753,9 @@ all handlers that match the rest."
 (defun iiimcf-im-destroyic-reply-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-trigger-notify-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-trigger-notify-handler-default)
  '(iiimp-im-trigger-notify)
@@ -762,7 +764,9 @@ all handlers that match the rest."
 (defun iiimcf-im-trigger-notify-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-trigger-notify-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-trigger-notify-handler-default)
  '(iiimp-im-trigger-notify)
@@ -771,7 +775,8 @@ all handlers that match the rest."
 (defun iiimcf-im-setimvalues-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-setimvalues-reply
-   (nth 1 mes) (nth 2 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-setimvalues-handler-default)
  '(iiimp-im-setimvalues)
@@ -781,7 +786,9 @@ all handlers that match the rest."
   (iiimp-add-debug-log "REPLY:\n")
   (iiimp-send-message
    'iiimp-im-forward-event-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-forward-event-handler-default)
  '(iiimp-im-forward-event)
@@ -792,7 +799,10 @@ all handlers that match the rest."
 (defun iiimcf-im-forward-event-with-operations-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-forward-event-with-operations-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes) (nth 5 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)
+   (iiimp-message-forward-event-operations mes)))
 (iiimcf-register-handler
  (function iiimcf-im-forward-event-handler-default)
  '(iiimp-im-forward-event-with-operations)
@@ -801,7 +811,9 @@ all handlers that match the rest."
 (defun iiimcf-im-preedit-start-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-preedit-start-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-preedit-start-handler-default)
  '(iiimp-im-preedit-start)
@@ -810,7 +822,9 @@ all handlers that match the rest."
 (defun iiimcf-im-preedit-draw-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-preedit-draw-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-preedit-draw-handler-default)
  '(iiimp-im-preedit-draw)
@@ -819,7 +833,9 @@ all handlers that match the rest."
 (defun iiimcf-im-preedit-done-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-preedit-done-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-preedit-done-handler-default)
  '(iiimp-im-preedit-done)
@@ -828,7 +844,9 @@ all handlers that match the rest."
 (defun iiimcf-im-status-start-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-status-start-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-status-start-handler-default)
  '(iiimp-im-status-start)
@@ -837,7 +855,9 @@ all handlers that match the rest."
 (defun iiimcf-im-status-draw-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-status-draw-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-status-draw-handler-default)
  '(iiimp-im-status-draw)
@@ -846,7 +866,9 @@ all handlers that match the rest."
 (defun iiimcf-im-status-done-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-status-done-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-status-done-handler-default)
  '(iiimp-im-status-done)
@@ -857,7 +879,9 @@ all handlers that match the rest."
 (defun iiimcf-im-lookup-choice-start-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-lookup-choice-start-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-lookup-choice-start-handler-default)
  '(iiimp-im-lookup-choice-start)
@@ -866,7 +890,9 @@ all handlers that match the rest."
 (defun iiimcf-im-lookup-choice-draw-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-lookup-choice-draw-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-lookup-choice-draw-handler-default)
  '(iiimp-im-lookup-choice-draw)
@@ -875,7 +901,9 @@ all handlers that match the rest."
 (defun iiimcf-im-lookup-choice-process-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-lookup-choice-process-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-lookup-choice-process-handler-default)
  '(iiimp-im-lookup-choice-process)
@@ -884,7 +912,9 @@ all handlers that match the rest."
 (defun iiimcf-im-lookup-choice-done-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-lookup-choice-done-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)))
 (iiimcf-register-handler
  (function iiimcf-im-lookup-choice-done-handler-default)
  '(iiimp-im-lookup-choice-done)
@@ -896,8 +926,11 @@ all handlers that match the rest."
 (defun iiimcf-im-aux-start-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-aux-start-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)
-   (nth 4 mes) (nth 5 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)
+   (iiimp-message-aux-window-id mes)
+   (iiimp-message-aux-im-name mes)))
 (iiimcf-register-handler
  (function iiimcf-im-aux-start-handler-default)
  '(iiimp-im-aux-start)
@@ -906,8 +939,11 @@ all handlers that match the rest."
 (defun iiimcf-im-aux-draw-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-aux-draw-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)
-   (nth 4 mes) (nth 5 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)
+   (iiimp-message-aux-window-id mes)
+   (iiimp-message-aux-im-name mes)))
 (iiimcf-register-handler
  (function iiimcf-im-aux-draw-handler-default)
  '(iiimp-im-aux-draw)
@@ -916,8 +952,11 @@ all handlers that match the rest."
 (defun iiimcf-im-aux-done-handler-default (mes)
   (iiimp-send-message
    'iiimp-im-aux-done-reply
-   (nth 1 mes) (nth 2 mes) (nth 3 mes)
-   (nth 4 mes) (nth 5 mes)))
+   (iiimp-message-com-id mes)
+   (iiimp-message-im-id mes)
+   (iiimp-message-ic-id mes)
+   (iiimp-message-aux-window-id mes)
+   (iiimp-message-aux-im-name mes)))
 (iiimcf-register-handler
  (function iiimcf-im-aux-done-handler-default)
  '(iiimp-im-aux-done)
