@@ -24,7 +24,7 @@
 ;;; Commentary:
 ;; This module supplies low level features for communications
 ;; via IIIMP.
-;; This library is dependent of Mule-UCS-Unicode modules.
+;; This library depends on Mule-UCS-Unicode modules.
 
 ;;; Code:
 
@@ -299,7 +299,50 @@
 (put 'iiimp-connection-error 'error-message
      "IIIMP: connection reset by peer.")
 
-;; COM-ID := (<PROCESS> <BUFFER> <HOST> <PORT> <END-POINT>)
+;; COM-ID := [<PROCESS> <BUFFER> <HOST> <PORT> <END-POINT>]
+
+(defsubst iiimp-com-id-process (com-id)
+  (aref com-id 0))
+
+(defsubst iiimp-com-id-buffer (com-id)
+  (aref com-id 1))
+
+(defsubst iiimp-com-id-host (com-id)
+  (aref com-id 2))
+
+(defsubst iiimp-com-id-port (com-id)
+  (aref com-id 3))
+
+(defsubst iiimp-com-id-end-point (com-id)
+  (aref com-id 4))
+
+(defsubst iiimp-record-end-point (com-id pt)
+  (aset com-id 4 pt)
+  com-id)
+
+;; IM-ID := (COM-ID id._NUMBER_of_IM)
+
+(defsubst iiimp-im-id-to-com-id (im-id)
+  (car im-id))
+
+(defsubst iiimp-im-id-to-handle (im-id)
+  (nth 1 im-id))
+
+(defsubst iiimp-construct-im-id (com-id im-num)
+  (list com-id im-num))
+
+;; IC-ID := (COM-ID id._NUMBER_of_IC id._NUMBER_of_IM)
+
+(defsubst iiimp-ic-id-to-com-id (ic-id)
+  (car ic-id))
+
+(defsubst iiimp-ic-id-to-handle (ic-id)
+  (nth 1 ic-id))
+
+(defsubst iiimp-construct-ic-id (com-id im-num ic-num)
+  (list com-id ic-num im-num))
+
+;;; network I/O
 
 (defun iiimp-open-network-channel (host port buf)
   (let ((coding-system-for-read nil)
@@ -321,23 +364,24 @@
   (let* ((buf (generate-new-buffer " *IIIMP*"))
 	 (proc (iiimp-open-network-channel host port buf))
 	 slot)
-    (setq slot (list proc buf host port
-		     (save-excursion
-		       (set-buffer buf)
-		       (point-min))))
+    ;; CREATE COM-ID
+    (setq slot (vector proc buf host port
+		       (save-excursion
+			 (set-buffer buf)
+			 (point-min))))
     (setq iiimp-communication-list
 	  (cons slot iiimp-communication-list))
     slot))
 
 (defun iiimp-destroy-network-channel (com-id)
-  (let* ((proc (car com-id))
+  (let* ((proc (iiimp-com-id-process com-id))
 	 (stat (process-status proc)))
     (while (not (or (eq stat 'closed)
 		    (eq stat nil)))
 	(delete-process proc)
 	(setq stat (process-status proc))
 	(sleep-for 1))
-    (kill-buffer (nth 1 com-id))
+    (kill-buffer (iiimp-com-id-buffer com-id))
     (setq iiimp-communication-list
 	  (delq com-id iiimp-communication-list))
     t))
@@ -347,7 +391,7 @@
 	   (car iiimp-async-invocation-handler)))
 
 (defun iiimp-enable-async-invocation (com-id func)
-  (let ((buf (nth 1 com-id)))
+  (let ((buf (iiimp-com-id-buffer com-id)))
     (save-excursion
       (set-buffer buf)
       (make-variable-buffer-local
@@ -366,27 +410,23 @@
 
 (defmacro iiimp-communication (com-id &rest form)
   `(save-excursion
-     (set-buffer (nth 1 ,com-id))
+     (set-buffer (iiimp-com-id-buffer ,com-id))
      (let ((after-change-functions nil))
        ,@form)))
 
 (defsubst iiimp-check-channel-connection (com-id)
-  (eq (process-status (car com-id)) 'open))
-
-(defsubst iiimp-record-end-point (com-id start)
-  (setcar (nthcdr 4 com-id) start)
-  com-id)
+  (eq (process-status (iiimp-com-id-process com-id)) 'open))
 
 (defun iiimp-check-channel-bytes (com-id start bytes wait)
   (if (null wait) (>= (point-max) (+ start bytes))
     (if (not (numberp wait)) (setq wait nil))
     (while (< (point-max) (+ start bytes))
-      (accept-process-output (car com-id) wait))
+      (accept-process-output (iiimp-com-id-process com-id) wait))
     t))
 
 (defun iiimp-wait-channel (com-id until)
   (while (< (point-max) until)
-    (accept-process-output (car com-id))))
+    (accept-process-output (iiimp-com-id-process com-id))))
 
 (defun iiimp-read-channel (com-id start min-bytes)
   (iiimp-wait-channel com-id (+ start min-bytes)))
@@ -402,7 +442,7 @@
   (if bytes
       (process-send-string (car com-id)
 			   (substring str 0 bytes))
-    (process-send-string (car com-id) str)))
+    (process-send-string (iiimp-com-id-process com-id) str)))
 
 (defun iiimp-packet-substring (packet start end)
   (if (stringp packet)
@@ -415,7 +455,7 @@
   (+ start bytes))
 
 (defun iiimp-packet-erase (packet start)
-  (cond ((listp packet)
+  (cond ((vectorp packet)
 	 (delete-region (point-min) start)
 	 (iiimp-record-end-point packet (point-min))
 	 (point-min))
@@ -587,6 +627,16 @@
 (defsubst iiimp-make-padding (n)
   (make-string (iiimp-padding-bytes n) 0))
 
+;; im,ic network representation.
+(defsubst iiimp-pack-im-ic (im-id &optional ic-id)
+  (concat
+   (iiimp-pack-number
+    (((iiimp-im-id-to-handle im-id) . 16)))
+   (if ic-id
+       (iiimp-pack-number
+	(((iiimp-im-id-to-handle ic-id) . 16))))))
+
+;; annotation id.
 (defsubst iiimp-aid-to-aid-spec (aid)
   (let ((ll iiimp-annotation-properties))
     (while (and ll (/= aid (nth 1 (car ll))))
@@ -595,6 +645,7 @@
 	;;(error "Invalid aid:%d" aid)
 	)))
 
+;; attributes.
 (defsubst iiimp-request-attrs-pack (syms)
   (let ((result "") val)
     (while syms
@@ -1493,9 +1544,27 @@
 	  start)))
 
 ;;;
-;;; Protocol Parser
+;;; IIIMP message.
 ;;;
 
+;; (<opcode> <com-id> <im-id> <ic-id or nil> [<message-specific-data>])
+;; |---------message common header(CPART)---|-message speicific part-|
+
+(defsubst iiimp-message-opcode (mes)
+  (car mes))
+
+(defsubst iiimp-message-com-id (mes)
+  (nth 1 mes))
+
+(defsubst iiimp-message-im-id (mes)
+  (nth 2 mes))
+
+(defsubst iiimp-message-ic-id (mes)
+  (nth 3 mes))
+
+(defsubst iiimp-message-specific-data (mes)
+  (nthcdr 4 mes))
+
 (defun iiimp-show-message (com-id opcode start len)
   (if iiimp-debug-buffer
       (let* ((end (+ start len))
@@ -1507,6 +1576,10 @@
 	  (iiimp-dump-binary str)
 	  (insert "]\n")))))
 
+;;;
+;;; Protocol Parser
+;;;
+
 (defsubst iiimp-parse-header (com-id start)
   (let (type opcode len)
     (setq start
@@ -1523,24 +1596,31 @@
     (cons (list opcode len) start)))
 
 (defsubst iiimp-parse-im-ic-reply-message-internal (com-id opcode start len)
-  (let (im-id ic-id)
+  (let (im-num ic-num)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16))
+	   ((im-num . 16)
+	    (ic-num . 16))
 	   com-id start))
-    (list opcode com-id im-id ic-id)))
+    (list opcode com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num))))
 
 (defun iiimp-parse-default (com-id opcode start len)
   (if iiimp-debug-flag
       (iiimp-show-message com-id opcode start len))
   nil)
 
+;; IM_CONNECT_REPLY => (<CPART> <language-list>)
+
+(defsubst iiimp-message-im-language-list (mes)
+  (nth 4 mes))
+
 (defun iiimp-parse-im-connect-reply (com-id opcode start len)
-  (let (slot im-id lln)
+  (let (slot im-num lln)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
+	   ((im-num . 16)
 	    (len . 16))
 	   com-id start))
     (setq slot
@@ -1549,22 +1629,39 @@
 	   (function iiimp-string-unpack))
 	  lln (car slot)
 	  start (cdr slot))
-    (list 'iiimp-im-connect-reply com-id im-id lln)))
+    (list 'iiimp-im-connect-reply
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  nil ; IC is not specified.
+	  lln)))
+
+;; IM_DISCONNECT_REPLY => (<CPART>)
 
 (defun iiimp-parse-im-disconnect-reply (com-id opcode start len)
-  (let (im-id dum)
+  (let (im-num dum)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
+	   ((im-num . 16)
 	    (dum . 16))
 	   com-id start))
-    (list 'iiimp-im-disconnect-reply com-id im-id)))
+    (list 'iiimp-im-disconnect-reply
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  nil)))
+
+;; IM_REGISTER_TRIGGER_KEYS => (<CPART> <keyevent-list(on)>
+;;                                      <keyevent-list(off)>)
+
+(defsubst iiimp-message-trigger-on-key-list (mes)
+  (nth 4 mes))
+(defsubst iiimp-message-trigger-off-key-list (mes)
+  (nth 5 mes))
 
 (defun iiimp-parse-im-register-trigger-keys (com-id opcode start len)
-  (let (slot im-id dum len kls-on kls-off)
+  (let (slot im-num dum len kls-on kls-off)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
+	   ((im-num . 16)
 	    (dum . 16)
 	    (len . 32))
 	   com-id start))
@@ -1585,35 +1682,52 @@
 	  kls-off (car slot)
 	  start (cdr slot))
     (list 'iiimp-im-register-trigger-keys
-	  com-id im-id kls-on kls-off)))
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  nil ; IC is not specified.
+	  kls-on kls-off)))
+
+;; IM_TRIGGER_NOTIFY => (<CPART> <flag>)
+
+(defsubst iiimp-message-trigger-notify-flag (mes)
+  (nth 4 mes))
 
 (defun iiimp-parse-im-trigger-notify (com-id opcode start len)
-  (let (im-id ic-id flag dum)
+  (let (im-num ic-num flag dum)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (flag . 16)
 	    (dum . 16))
 	   com-id start))
     (list 'iiimp-im-trigger-notify
-	  com-id im-id ic-id
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
 	  (cond ((= flag 0) t)
 		((= flag 1) nil)
 		(t (error "IIIM Protocol error:IM_TRIGGER_NOTIFY(flag):%S"
 			  flag))))))
 
+;; IM_TRIGGER_NOTIFY_REPLY => (<CPART>)
+
 (defun iiimp-parse-im-trigger-notify-reply (com-id opcode start len)
   (iiimp-parse-im-ic-reply-message-internal
    com-id opcode start len))
 
 ;;; get/set IM values
 
+;; IM_GET/SET_IMVALUES => (<CPART> <im-attribtute-list>)
+
+(defsubst iiimp-message-im-attribute-list (mes)
+  (nth 4 mes))
+
 (defun iiimp-parse-imvalues-internal (com-id opcode start len)
-  (let (slot im-id dum lia)
+  (let (slot im-num dum lia)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
+	   ((im-num . 16)
 	    (dum . 16)
 	    (len . 32))
 	   com-id start))
@@ -1623,7 +1737,10 @@
 	   (function iiimp-im-attribute-unpack))
 	  lia (car slot)
 	  start (cdr slot))
-    (list opcode com-id im-id lia)))
+    (list opcode com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  nil ; IC is not specified.
+	  lia)))
 
 (defun iiimp-parse-im-getimvalues-reply (com-id opcode start len)
   (iiimp-parse-imvalues-internal com-id opcode start len))
@@ -1632,13 +1749,17 @@
   (iiimp-parse-imvalues-internal com-id opcode start len))
 
 (defun iiimp-parse-im-setimvalues-reply (com-id opcode start len)
-  (let (im-id dum)
+  (let (im-num dum)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
+	   ((im-num . 16)
 	    (dum . 16))
 	   com-id start))
-    (list 'iiimp-im-setimvalues-reply com-id im-id)))
+    (list 'iiimp-im-setimvalues-reply
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  nil ; IC is not specified.
+	  )))
 
 ;;; IC messages
 
@@ -1654,12 +1775,17 @@
   (iiimp-parse-im-ic-reply-message-internal
    com-id opcode start len))
 
+;; IM_GETICVALUES_REPLY => (<CPART> <ic-attribute-list>)
+
+(defsubst iiimp-message-ic-attribute-list (mes)
+  (nth 4 mes))
+
 (defun iiimp-parse-im-geticvalues-reply (com-id opcode start len)
-  (let (slot im-id ic-id flag lia)
+  (let (slot im-num ic-num flag lia)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (len . 16))
 	   com-id start))
     (setq slot
@@ -1668,7 +1794,11 @@
 	   (function iiimp-icattribute-unpack))
 	  lia (car slot)
 	  start (cdr slot))
-    (list opcode com-id im-id lia)))
+    (list opcode
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  lia)))
 
 (defun iiimp-parse-im-seticfocus-reply (com-id opcode start len)
   (iiimp-parse-im-ic-reply-message-internal
@@ -1684,26 +1814,40 @@
 
 ;;; forward event
 
+;; IM_FORWARD_EVENT => (<CPART> <contents>)
+
+(defsubst iiimp-message-forward-event-contents (mes)
+  (nth 4 mes))
+
 (defun iiimp-parse-im-forward-event (com-id opcode start len)
-  (let (im-id ic-id contents)
+  (let (im-num ic-num contents)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16))
+	   ((im-num . 16)
+	    (ic-num . 16))
 	   com-id start))
     (setq contents (car (iiimp-contents-unpack com-id start)))
-    (list 'iiimp-im-forward-event com-id im-id ic-id contents)))
+    (list 'iiimp-im-forward-event
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  contents)))
 
 (defun iiimp-parse-im-forward-event-reply (com-id opcode start len)
   (iiimp-parse-im-ic-reply-message-internal
    com-id opcode start len))
 
+;; IM_FORWARD_EVENT_WITH_OPERATIONS => (<CPART> <contents> <operations>)
+
+(defsubst iiimp-message-forward-event-operations (mes)
+  (nth 5 mes))
+
 (defun iiimp-parse-im-forward-event-with-operations (com-id opcode start len)
-  (let (im-id ic-id contents len ops)
+  (let (im-num ic-num contents len ops)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16))
+	   ((im-num . 16)
+	    (ic-num . 16))
 	   com-id start))
     (setq contents (iiimp-contents-unpack com-id start)
 	  start (cdr contents)
@@ -1717,15 +1861,22 @@
 		com-id start len
 		(function iiimp-operation-unpack))))
     (list 'iiimp-im-forward-event-with-operations
-	  com-id im-id ic-id contents ops)))
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  contents
+	  ops)))
+
+;; IM_FORWARD_EVENT_WITH_OPERATIONS_REPLY
+;;            => (<CPART> nil <operations>)
 
 (defun iiimp-parse-im-forward-event-with-operations-reply
   (com-id opcode start len)
-  (let (im-id ic-id len ops)
+  (let (im-num ic-num len ops)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (len . 32))
 	   com-id start))
     (setq ops
@@ -1733,37 +1884,62 @@
 		com-id start len
 		(function iiimp-operation-unpack))))
     (list 'iiimp-im-forward-event-with-operations-reply
-	  com-id im-id ic-id ops)))
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  nil
+	  ops)))
+
+;; IM_COMMIT_STRING => (<CPART> <contents>)
+
+(defsubst iiimp-message-committed-string (mes)
+  (nth 4 mes))
 
 (defun iiimp-parse-im-commit-string (com-id opcode start len)
-  (let (im-id ic-id contents)
+  (let (im-num ic-num contents)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16))
+	   ((im-num . 16)
+	    (ic-num . 16))
 	   com-id start))
     (setq contents (car (iiimp-contents-unpack com-id start)))
-    (list 'iiimp-im-commit-string com-id im-id ic-id contents)))
+    (list 'iiimp-im-commit-string
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  contents)))
+
+;; IM_PREEDIT_START => (<CPART>)
 
 (defun iiimp-parse-im-preedit-start (com-id opcode start len)
   (iiimp-parse-im-ic-reply-message-internal
    com-id opcode start len))
 
+;; IM_PREEDIT_DRAW => (<CPART>
+;;                     [<caretpos> <chg-st> <chg-len> <contents>])
+
+;; ensure the result is [<caretpos> <chg-st> <chg-len> <contents>].
+(defsubst iiimp-message-preedit-draw-data (mes)
+  (nth 4 mes))
+
 (defun iiimp-parse-im-preedit-draw (com-id opcode start len)
-  (let (im-id ic-id caret chg-first chg-len contents)
+  (let (im-num ic-num caret chg-first chg-len contents)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (caret . 32)
 	    (chg-first . 32)
 	    (chg-len . 32))
 	   com-id start))
     (setq contents (car (iiimp-contents-unpack com-id start)))
     (list 'iiimp-im-preedit-draw
-	  com-id im-id ic-id
-	  caret chg-first chg-len
-	  contents)))
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  (vector caret chg-first chg-len contents)
+	  ;; message ends here.
+	  )))
 
 (defun iiimp-parse-im-preedit-done (com-id opcode start len)
   (iiimp-parse-im-ic-reply-message-internal
@@ -1773,15 +1949,24 @@
   (iiimp-parse-im-ic-reply-message-internal
    com-id opcode start len))
 
+;; IM_STATUS_DRAW => (<CPART> <contents>)
+
+(defsubst iiimp-message-status-draw-contents (mes)
+  (nth 4 mes))
+
 (defun iiimp-parse-im-status-draw (com-id opcode start len)
-  (let (im-id ic-id contents)
+  (let (im-num ic-num contents)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16))
+	   ((im-num . 16)
+	    (ic-num . 16))
 	   com-id start))
     (setq contents (car (iiimp-contents-unpack com-id start)))
-    (list 'iiimp-im-status-draw com-id im-id ic-id contents)))
+    (list 'iiimp-im-status-draw
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  contents)))
 
 (defun iiimp-parse-im-status-done (com-id opcode start len)
   (iiimp-parse-im-ic-reply-message-internal
@@ -1789,13 +1974,22 @@
 
 ;;; LOOKUP messages.
 
+;; IM_LOOKUP_CHOICE_START =>
+;;      (<CPART>
+;;       [<masterp> <choices-num> <rows> <cols> <direction> <labelownp>])
+
+;; ensure the result is (<masterp> <choices-num> <rows> <cols>
+;;                       <direction> <labelownp>)
+(defsubst iiimp-message-lookup-choice-start-data (mes)
+  (nth 4 mes))
+
 (defun iiimp-parse-im-lookup-choice-start (com-id opcode start len)
-  (let (im-id ic-id masterp choices-num
+  (let (im-num ic-num masterp choices-num
 	rows cols dir labelownp)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (masterp . 16)
 	    (choices-num . 16)
 	    (rows . 16)
@@ -1804,21 +1998,35 @@
 	    (labelownp . 16))
 	   com-id start))
     (setq masterp (= masterp 2))
-    (setq dir (if (= dir 0) 'vertical 'horizontal))
+    (setq dir (if (= dir 0) 'horizontal 'vertical))
     (setq labelownp (if (= labelownp 0) nil t))
     (list 'iiimp-im-lookup-choice-start
-	  com-id im-id ic-id masterp choices-num
-	  rows cols dir labelownp)))
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  (vector masterp choices-num rows cols dir labelownp)
+	  ;; message ends here
+	  )))
+
+;; IM_LOOKUP_CHOICE_DRAW =>
+;;      (<CPART>
+;;       [<idxfirst> <idxlast> <idxcurrent>
+;;        <choice-list> <label-list> <title>])
+
+;; ensure the result is [<idxfirst> <idxlast> <idxcurrent>
+;;                       <choice-list> <label-list> <title>]
+(defsubst iiimp-message-lookup-choice-draw-data (mes)
+  (nth 4 mes))
 
 (defun iiimp-parse-im-lookup-choice-draw (com-id opcode start len)
-  (let (slot im-id ic-id
+  (let (slot im-num ic-num
 	idxfirst idxlast idxcur
 	chlist chlistlen lblist lblistlen
 	title)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (idxfirst . 32)
 	    (idxlast . 32)
 	    (idxcur . 32)
@@ -1840,16 +2048,28 @@
 	  start (cdr slot))
     (setq title (car (iiimp-text-unpack com-id start)))
     (list 'iiimp-im-lookup-choice-draw
-	  com-id im-id ic-id
-	  idxfirst idxlast idxcur
-	  chlist lblist title)))
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  (vector idxfirst idxlast idxcur
+		  chlist lblist title)
+	  ;; message ends here.
+	  )))
+
+;; IM_LOOKUP_CHOICE_PROCESS => (<CPART> <type> <value>)
+
+(defsubst iiimp-message-lookup-choice-process-type (mes)
+  (nth 4 mes))
+
+(defsubst iiimp-message-lookup-choice-process-value (mes)
+  (nth 5 mes))
 
 (defun iiimp-parse-im-lookup-choice-process (com-id opcode start len)
-  (let (im-id ic-id type val)
+  (let (im-num ic-num type val)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (type . 16)
 	    (val . 16))
 	   com-id start))
@@ -1865,7 +2085,10 @@
 	    (t
 	     (setq val 'last))))
     (list 'iiimp-im-lookup-choice-process
-	  com-id im-id ic-id type val)))
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  type val)))
 
 (defun iiimp-parse-im-lookup-choice-done (com-id opcode start len)
   (iiimp-parse-im-ic-reply-message-internal
@@ -1873,22 +2096,34 @@
 
 ;;; AUX messages.
 
+;; IM_AUX_* => (<CPART> <window-id> <im-name>
+;;                      [<specific data of aux-draw>])
+
+(defsubst iiimp-message-aux-window-id (mes)
+  (nth 4 mes))
+
+(defsubst iiimp-message-aux-im-name (mes)
+  (nth 5 mes))
+
 ;; (iiimp-im-aux-start
 ;;   <COM-ID> <IM-ID> <IC-ID>
 ;;   (<WindowClassID1> . <WindowClassID2>)
 ;;   <STR:InputMethodName>)
+
 (defun iiimp-parse-im-aux-start (com-id opcode start len)
-  (let (im-id ic-id wid1 wid2 im-name)
+  (let (im-num ic-num wid1 wid2 im-name)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (wid1 . 16)
 	    (wid2 . 16))
 	   com-id start))
     (setq im-name (car (iiimp-string-unpack com-id start)))
     (list 'iiimp-im-aux-start
-	  com-id im-id ic-id
+	  com-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
 	  (cons wid1 wid2)
 	  im-name)))
 
@@ -1898,13 +2133,20 @@
 ;;   <STR:InputMethodName>
 ;;   ((NUM-MSW . NUM-LSW) ...)
 ;;   (STR STR ...))
+
+(defsubst iiimp-message-aux-int-list (mes)
+  (nth 6 mes))
+
+(defsubst iiimp-message-aux-string-list (mes)
+  (nth 7 mes))
+
 (defun iiimp-parse-im-aux-draw (com-id opcode start len)
-  (let (im-id ic-id wid1 wid2 slot im-name
+  (let (im-num ic-num wid1 wid2 slot im-name
 	num1 num2 len ivals svals)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (wid1 . 16)
 	    (wid2 . 16))
 	   com-id start))
@@ -1936,7 +2178,9 @@
 		      (function iiimp-string-unpack))))
     (list 'iiimp-im-aux-draw
 	  com-id
-	  im-id ic-id (cons wid1 wid2)
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
+	  (cons wid1 wid2)
 	  im-name ivals svals)))
   
 ;; (iiimp-im-aux-done
@@ -1944,18 +2188,19 @@
 ;;   (<WindowClassID1> . <WindowClassID2>)
 ;;   <STR:InputMethodName>)
 (defun iiimp-parse-im-aux-done (com-id opcode start len)
-  (let (im-id ic-id wid1 wid2 im-name)
+  (let (im-num ic-num wid1 wid2 im-name)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (wid1 . 16)
 	    (wid2 . 16))
 	   com-id start))
     (setq im-name (car (iiimp-string-unpack com-id start)))
     (list 'iiimp-im-aux-done
 	  com-id
-	  im-id ic-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
 	  (cons wid1 wid2)
 	  im-name)))
 
@@ -1964,18 +2209,19 @@
 ;;   (<WindowClassID1> . <WindowClassID2>)
 ;;   <STR:InputMethodName>)
 (defun iiimp-parse-im-aux-setvalues-reply (com-id opcode start len)
-  (let (im-id ic-id wid1 wid2 im-name)
+  (let (im-num ic-num wid1 wid2 im-name)
     (setq start
 	  (iiimp-unpack-number
-	   ((im-id . 16)
-	    (ic-id . 16)
+	   ((im-num . 16)
+	    (ic-num . 16)
 	    (wid1 . 16)
 	    (wid2 . 16))
 	   com-id start))
     (setq im-name (car (iiimp-string-unpack com-id start)))
     (list 'iiimp-im-aux-setvalues-reply
 	  com-id
-	  im-id ic-id
+	  (iiimp-construct-im-id com-id im-num)
+	  (iiimp-construct-ic-id com-id im-num ic-num)
 	  (cons wid1 wid2)
 	  im-name)))
 
@@ -1992,7 +2238,7 @@
 (defun iiimp-wait-message (com-id
 			   &optional
 			   wait intercept-message not-discard)
-  (let ((start (nth 4 com-id))
+  (let ((start (iiimp-com-id-end-point com-id))
 	st2 type opcode len parser val)
     (iiimp-communication
      com-id
@@ -2031,7 +2277,7 @@
    com-id
    (iiimp-make-header
     opcode
-    (iiimp-pack-number ((im-id . 16) (ic-id . 16))))))
+    (iiimp-pack-im-ic im-id ic-id))))
 
 (defun iiimp-send-im-connect (com-id username &optional authlist)
   (let ((authlistpack (iiimp-pack-list
@@ -2052,20 +2298,22 @@
    com-id
    (iiimp-make-header
     'iiimp-im-disconnect
-    (iiimp-pack-number ((im-id . 16) (0 . 16))))))
+    (concat
+     (iiimp-pack-im-ic im-id)
+     (iiimp-pack-number ((0 . 16)))))))
 
 (defun iiimp-send-im-trigger-notify (com-id im-id ic-id flag)
   (iiimp-write-channel
    com-id
    (iiimp-make-header
     'iiimp-im-trigger-notify
-    (if flag
+    (concat
+     (iiimp-pack-im-ic im-id ic-id)
+     (if flag
 	(iiimp-pack-number
-	 ((im-id . 16) (ic-id . 16)
-	  (0 . 16) (0 . 16)))
+	 ((0 . 16) (0 . 16)))
       (iiimp-pack-number
-       ((im-id . 16) (ic-id . 16)
-	(1 . 16) (0 . 16)))))))
+       ((1 . 16) (0 . 16))))))))
 
 (defun iiimp-send-im-trigger-notify-reply (com-id im-id ic-id)
   (iiimp-send-im-ic-request-internal
@@ -2081,9 +2329,9 @@
      (iiimp-make-header
       'iiimp-im-setimvalues
       (concat
+       (iiimp-pack-im-ic im-id)
        (iiimp-pack-number
-	((im-id . 16)
-	 (0 . 16) ;; for padding
+	((0 . 16) ;; for padding
 	 (len . 32)))
        packed)))))
 
@@ -2092,10 +2340,10 @@
    com-id
    (iiimp-make-header
     'iiimp-im-setimvalues-reply
-    (iiimp-pack-number
-     ((im-id . 16)
-      (0 . 16)) ;; for padding
-     ))))
+    (concat
+     (iiimp-pack-im-ic im-id)
+     (iiimp-pack-number ((0 . 16)) ;; for padding
+     )))))
 
 (defun iiimp-send-im-getimvalues (com-id im-id request-attrs)
   (let* ((packed (iiimp-request-attrs-pack
@@ -2106,9 +2354,9 @@
      (iiimp-make-header
       'iiimp-im-getimvalues
       (concat
+       (iiimp-pack-im-ic im-id)
        (iiimp-pack-number
-	((im-id . 16)
-	 (0 . 16) ;; for padding
+	((0 . 16) ;; for padding
 	 (len . 32)))
        packed
        (iiimp-make-padding len))))))
@@ -2123,9 +2371,9 @@
      (iiimp-make-header
       'iiimp-im-createic
       (concat
+       (iiimp-pack-im-ic im-id)
        (iiimp-pack-number
-	((im-id . 16)
-	 (len . 16)))
+	((len . 16)))
        packed)))))
 
 (defun iiimp-send-im-destroyic (com-id im-id ic-id)
@@ -2142,10 +2390,8 @@
      (iiimp-make-header
       'iiimp-im-seticvalues
       (concat
-       (iiimp-pack-number
-	((im-id . 16)
-	 (ic-id . 16)
-	 (len . 16)))
+       (iiimp-pack-im-ic im-id ic-id)
+       (iiimp-pack-number ((len . 16)))
        packed
        (iiimp-make-padding 2))))))
 
@@ -2158,10 +2404,9 @@
      (iiimp-make-header
       'iiimp-im-geticvalues
       (concat
+       (iiimp-pack-im-ic im-id ic-id)
        (iiimp-pack-number
-	((im-id . 16)
-	 (ic-id . 16) ;; for padding
-	 (len . 16)))
+	((len . 16)))
        packed
        (iiimp-make-padding (+ 2 len)))))))
 
@@ -2183,7 +2428,7 @@
    (iiimp-make-header
     'iiimp-im-forward-event
     (concat
-     (iiimp-pack-number ((im-id . 16) (ic-id . 16)))
+     (iiimp-pack-im-ic im-id ic-id)
      (iiimp-contents-pack contents)))))
 
 (defun iiimp-send-im-forward-event-reply (com-id im-id ic-id)
@@ -2201,7 +2446,7 @@
    (iiimp-make-header
     'iiimp-im-forward-event-with-operations
     (concat
-     (iiimp-pack-number ((im-id . 16) (ic-id . 16)))
+     (iiimp-pack-im-ic im-id ic-id)
      (iiimp-contents-pack contents)
      (iiimp-pack-number ((len . 32)))
      packed)))))
@@ -2217,7 +2462,8 @@
    (iiimp-make-header
     'iiimp-im-forward-event-with-operations
     (concat
-     (iiimp-pack-number ((im-id . 16) (ic-id . 16) (len . 32)))
+     (iiimp-pack-im-ic im-id ic-id)
+     (iiimp-pack-number ((len . 32)))
      packed)))))
 
 (defun iiimp-send-im-preedit-start-reply
@@ -2226,11 +2472,11 @@
    com-id
    (iiimp-make-header
     'iiimp-im-preedit-start-reply
-    (if (null maxlen)
-	(iiimp-pack-number ((im-id . 16) (ic-id . 16)
-			    (65535 . 16) (65535 . 16)))
-      (iiimp-pack-number ((im-id . 16) (ic-id . 16)
-			  (maxlen . 32)))))))
+    (concat
+     (iiimp-pack-im-ic im-id ic-id)
+     (if (null maxlen)
+	 (iiimp-pack-number ((65535 . 16) (65535 . 16)))
+       (iiimp-pack-number ((maxlen . 32))))))))
 
 (defun iiimp-send-im-preedit-draw-reply (com-id im-id ic-id)
   (iiimp-send-im-ic-request-internal
@@ -2238,7 +2484,7 @@
 
 (defun iiimp-send-im-preedit-done-reply (com-id im-id ic-id)
   (iiimp-send-im-ic-request-internal
-   com-id 'iiimp-im-preedit-draw-reply im-id ic-id))
+   com-id 'iiimp-im-preedit-done-reply im-id ic-id))
 
 (defun iiimp-send-im-status-start-reply (com-id im-id ic-id)
   (iiimp-send-im-ic-request-internal
@@ -2279,9 +2525,9 @@
    (iiimp-make-header
     'iiimp-im-aux-start-reply
     (concat
+     (iiimp-pack-im-ic im-id ic-id)
      (iiimp-pack-number
-      ((im-id . 16) (ic-id . 16)
-       ((car wid) . 16) ((cdr wid) . 16)))
+      (((car wid) . 16) ((cdr wid) . 16)))
      (iiimp-string-pack im-name)))))
 
 (defun iiimp-send-im-aux-draw-reply
@@ -2291,9 +2537,9 @@
    (iiimp-make-header
     'iiimp-im-aux-draw-reply
     (concat
+     (iiimp-pack-im-ic im-id ic-id)
      (iiimp-pack-number
-      ((im-id . 16) (ic-id . 16)
-       ((car wid) . 16) ((cdr wid) . 16)))
+      (((car wid) . 16) ((cdr wid) . 16)))
      (iiimp-string-pack im-name)))))
 
 (defun iiimp-send-im-aux-done-reply
@@ -2303,9 +2549,9 @@
    (iiimp-make-header
     'iiimp-im-aux-done-reply
     (concat
+     (iiimp-pack-im-ic im-id ic-id)
      (iiimp-pack-number
-      ((im-id . 16) (ic-id . 16)
-       ((car wid) . 16) ((cdr wid) . 16)))
+      (((car wid) . 16) ((cdr wid) . 16)))
      (iiimp-string-pack im-name)))))
 
 (defun iiimp-send-im-aux-setvalues
@@ -2322,9 +2568,9 @@
    (iiimp-make-header
     'iiimp-im-aux-setvalues
     (concat
+     (iiimp-pack-im-ic im-id ic-id)
      (iiimp-pack-number
-      ((im-id . 16) (ic-id . 16)
-       ((car wid) . 16) ((cdr wid) . 16)))
+      (((car wid) . 16) ((cdr wid) . 16)))
      (iiimp-string-pack im-name)
      (iiimp-pack-number
       ((dwlen . 32)))
